/*************************************************************************
*                                                                        *
*   Module: JoinGenes                                                    *
*                                                                        *
*   Join best genes of each sequence split                               *
*                                                                        *
*   This file is part of the geneidP Distribution                        *
*                                                                        *
*     Copyright (C) 2000 - Enrique BLANCO GARCIA                         *
*                          Roderic GUIGO SERRA                           *
*                                                                        *
*  This program is free software; you can redistribute it and/or modify  *
*  it under the terms of the GNU General Public License as published by  *
*  the Free Software Foundation; either version 2 of the License, or     *
*  (at your option) any later version.                                   *
*                                                                        *
*  This program is distributed in the hope that it will be useful,       *
*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*  GNU General Public License for more details.                          *
*                                                                        *
*  You should have received a copy of the GNU General Public License     *
*  along with this program; if not, write to the Free Software           *
*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.             *
*************************************************************************/

/* $Id: JoinGenes.c,v 1.1 2001/01/08 08:08:42 eblanco Exp $ */

#include "geneid.h"  

/* Equality of exons */
int equalExons(exonGFF* exonL, exonGFF* exonR)
{
  int equal;

  if ( ((exonR->Acceptor->Position + exonR->offset1) ==
	(exonL->Acceptor->Position + exonL->offset1)) &&
       ((exonR->Donor->Position + exonR->offset2) ==
	(exonL->Donor->Position + exonL->offset2))  &&
       (exonR->Strand == exonL->Strand) &&
       (exonR->Frame == exonL->Frame) &&
       (!strcmp(exonR->Type,exonL->Type)))
    equal = 1;
  else
    equal = 0;

  return(equal);
}

/* Recomputing score of gene result from an operation of join of two genes */
double RecomputeGeneScore(exonGFF* currExon, exonGFF* joinExon)
{
  double scoreAcum;

  /* down recursive call */
  if (equalExons(currExon,joinExon))
    {
      currExon->GeneScore = joinExon->GeneScore;
      scoreAcum = joinExon->GeneScore;
    }
  /* up recursive call */
  else
    {
      scoreAcum = RecomputeGeneScore(currExon->PreviousExon,joinExon) + currExon->Score;
      currExon->GeneScore = scoreAcum;
    }
  return(scoreAcum);
}

/* Join two best genes. Result will be at genR */
int Join2Genes(exonGFF* genL, exonGFF* genR)
{
  exonGFF* exonL;
  exonGFF* exonR;
  int eogL, eogR, joined;
  int res;

  joined = 0;
  exonL  = genL;
  exonR  = genR;
  eogL = (exonL->Strand == '*');
  eogR = (exonR->Strand == '*');

  while((!joined) && (!eogL) && (!eogR))
    {
      if ( (exonR->Acceptor->Position + exonR->offset1) >
	   (exonL->Acceptor->Position + exonL->offset1) )
	{
	  exonR = exonR->PreviousExon;
	}
      else
	if ( (exonR->Acceptor->Position + exonR->offset1) <
	     (exonL->Acceptor->Position + exonL->offset1) )
	  {
	    exonL = exonL->PreviousExon;
	  }
	else
	  if (equalExons(exonL,exonR))
	    joined = 1;
	  else
	    {
	      /* The same acceptor but different exons */
	      exonL = exonL->PreviousExon;
	      exonR = exonR->PreviousExon;
	    }

      eogL = (exonL->Strand == '*');
      eogR = (exonR->Strand == '*');
    }

  if (joined)
    {
      /* Joinning the genes */
      exonR->PreviousExon = exonL->PreviousExon;
      RecomputeGeneScore(genR,exonL);
      res = 1;
    }
  else
    res = 0;

  return res;
}

void RecomputeGene(params*p1, params*p2)
{
  int res,i;
  void* status;
  pthread_t t;
  char mess[MAXSTRING];
  int aux, aux2;

  /* Best gene will be at p2 */
  p2->quiensoy = NUMTHREADS;

  /* Reset params */
  p2->left = 0;
  p2->right = p2->right;

  for(i=0; i < MAXENTRY; i++)
    {    
      p2->genes->km[i] = 0;
      p2->genes->je[i] = 0;
    }

  /* Mark exon as Ghost Exon */
  p2->genes->Ghost -> Strand = '*';
  p2->genes->Ghost -> GeneScore = 0.0;
  p2->genes->Ghost -> Donor -> Position = 0;
  p2->genes->Ghost -> offset1 = 0;
  p2->genes->Ghost -> offset2 = 0;
  p2->genes->GOptim = p2->genes->Ghost;
  
  for(aux=0; aux<MAXENTRY; aux++)
    {
      for(aux2=0; aux2 < FRAMES; aux2++)
        p2->genes->Ga[aux][aux2] = p2->genes->Ghost;
    } 

  p2->dumpster->ndumpSites = 0;
  p2->dumpster->ndumpExons = 0;
  resetDumpHash(p2->dumpster->h);  

  p2->i1vExons = 0;
  p2->i2vExons = 0;

  /* Creating One thread */
  res = pthread_create(&t,NULL,
		       (void*)ThreadWork,(void*) p2);
  if (res != 0)
    {
      sprintf(mess,"Pthread_create: %s\n",strerror(res));
      printError(mess);
    }
  
  /* Waiting One thread */
  res = pthread_join(t,&status);
  if (res != 0)
    {
      sprintf(mess,"Pthread_join: %s\n",strerror(res));
      printError(mess);
    }
  if (status != NULL)
    {
      sprintf(mess,"Wrong finish: Pthread %d !!\n",p2->quiensoy);
      printError(mess);
    }
  else
    {
      sprintf(mess,"Right finish: Pthread %d \n",p2->quiensoy);
      printMess(mess);
    }
}

/* Control process of joinning of N genes */
void JoinGenes(params* p[NUMTHREADS])
{
  int j;
  int res;
  char mess[MAXSTRING];

  for(j=1; j<NUMTHREADS; j++)
    {
      res = Join2Genes(p[j-1]->genes->GOptim, p[j]->genes->GOptim);
      if (res == 0)
	{
	  sprintf(mess,"Recomputing genes between threads %d - %d",j-1,j);
	  printMess(mess);
	  
	  /* Recompute best gene in this region */
	  RecomputeGene(p[j-1],p[j]);
	}
      else
	{
	  sprintf(mess,"Right join genes between threads %d - %d",j-1,j);
	  printMess(mess);
	}
    }
}






